#
# Copyright (C) 2019 Databricks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
A metadata to manage indexes.
"""

from typing import List, Optional, Tuple

import pandas as pd

from databricks import koalas as ks  # For running doctests and reference resolution in PyCharm.


IndexMap = Tuple[str, Optional[str]]


class Metadata(object):
    """
    Manages column names and index information.

    :ivar _data_columns: list of the Spark field names to be seen as columns in Koalas DataFrame.
    :ivar _index_map: list of pair holding the Spark field names for indexes,
                       and the index name to be seen in Koalas DataFrame.

    .. note:: this is an internal class. It is not supposed to be exposed to users and users
        should not directly access to it.

    Metadata represents the index information for a DataFrame it belongs to. For instance,
    if we have a Koalas DataFrame as below, Pandas DataFrame does not store the index as columns.

    >>> kdf = ks.DataFrame({
    ...     'A': [1, 2, 3, 4],
    ...     'B': [5, 6, 7, 8],
    ...     'C': [9, 10, 11, 12],
    ...     'D': [13, 14, 15, 16],
    ...     'E': [17, 18, 19, 20]}, columns = ['A', 'B', 'C', 'D', 'E'])
    >>> kdf  # doctest: +NORMALIZE_WHITESPACE
       A  B   C   D   E
    0  1  5   9  13  17
    1  2  6  10  14  18
    2  3  7  11  15  19
    3  4  8  12  16  20

    However, all columns including index column are also stored in Spark DataFrame internally
    as below.

    >>> kdf.to_spark().show()  # doctest: +NORMALIZE_WHITESPACE
    +-----------------+---+---+---+---+---+
    |__index_level_0__|  A|  B|  C|  D|  E|
    +-----------------+---+---+---+---+---+
    |                0|  1|  5|  9| 13| 17|
    |                1|  2|  6| 10| 14| 18|
    |                2|  3|  7| 11| 15| 19|
    |                3|  4|  8| 12| 16| 20|
    +-----------------+---+---+---+---+---+

    In order to fill this gap, the current metadata is used by mapping Spark's internal column
    to Koalas' index. See the method below:

    * `data_columns` represents non-indexing columns

    * `index_columns` represents internal index columns

    * `columns` represents all columns

    * `index_names` represents the external index name

    * `index_map` is zipped pairs of `index_columns` and `index_names`

    >>> metadata = kdf._metadata
    >>> metadata.data_columns
    ['A', 'B', 'C', 'D', 'E']
    >>> metadata.index_columns
    ['__index_level_0__']
    >>> metadata.columns
    ['__index_level_0__', 'A', 'B', 'C', 'D', 'E']
    >>> metadata.index_names
    [None]
    >>> metadata.index_map
    [('__index_level_0__', None)]

    In case that index is set to one of the existing column as below:

    >>> kdf1 = kdf.set_index("A")
    >>> kdf1  # doctest: +NORMALIZE_WHITESPACE
       B   C   D   E
    A
    1  5   9  13  17
    2  6  10  14  18
    3  7  11  15  19
    4  8  12  16  20

    >>> kdf1.to_spark().show()  # doctest: +NORMALIZE_WHITESPACE
    +---+---+---+---+---+
    |  A|  B|  C|  D|  E|
    +---+---+---+---+---+
    |  1|  5|  9| 13| 17|
    |  2|  6| 10| 14| 18|
    |  3|  7| 11| 15| 19|
    |  4|  8| 12| 16| 20|
    +---+---+---+---+---+

    >>> metadata = kdf1._metadata
    >>> metadata.data_columns
    ['B', 'C', 'D', 'E']
    >>> metadata.index_columns
    ['A']
    >>> metadata.columns
    ['A', 'B', 'C', 'D', 'E']
    >>> metadata.index_names
    ['A']
    >>> metadata.index_map
    [('A', 'A')]

    In case that index becomes a multi index as below:

    >>> kdf2 = kdf.set_index("A", append=True)
    >>> kdf2  # doctest: +NORMALIZE_WHITESPACE
         B   C   D   E
      A
    0 1  5   9  13  17
    1 2  6  10  14  18
    2 3  7  11  15  19
    3 4  8  12  16  20

    >>> kdf2.to_spark().show()  # doctest: +NORMALIZE_WHITESPACE
    +-----------------+---+---+---+---+---+
    |__index_level_0__|  A|  B|  C|  D|  E|
    +-----------------+---+---+---+---+---+
    |                0|  1|  5|  9| 13| 17|
    |                1|  2|  6| 10| 14| 18|
    |                2|  3|  7| 11| 15| 19|
    |                3|  4|  8| 12| 16| 20|
    +-----------------+---+---+---+---+---+

    >>> metadata = kdf2._metadata
    >>> metadata.data_columns
    ['B', 'C', 'D', 'E']
    >>> metadata.index_columns
    ['__index_level_0__', 'A']
    >>> metadata.columns
    ['__index_level_0__', 'A', 'B', 'C', 'D', 'E']
    >>> metadata.index_names
    [None, 'A']
    >>> metadata.index_map
    [('__index_level_0__', None), ('A', 'A')]
    """

    def __init__(self, data_columns: List[str],
                 index_map: Optional[List[IndexMap]] = None) -> None:
        """ Create a new metadata to manage column fields and index fields and names.

        :param data_columns: list of string
                              Field names to appear as columns.
        :param index_map: list of string pair
                           Each pair holds the index field name which exists in Spark fields,
                           and the index name.
        """
        assert all(isinstance(col, str) for col in data_columns)
        assert index_map is None \
            or all(isinstance(index_field, str)
                   and (index_name is None or isinstance(index_name, str))
                   for index_field, index_name in index_map)
        self._data_columns = data_columns  # type: List[str]
        self._index_map = index_map or []  # type: List[IndexMap]

    @property
    def data_columns(self) -> List[str]:
        """ Returns the managed column field names. """
        return self._data_columns

    @property
    def index_columns(self) -> List[str]:
        """ Returns the managed index field names. """
        return [index_column for index_column, _ in self._index_map]

    @property
    def columns(self) -> List[str]:
        """ Return all the field names including index field names. """
        index_columns = self.index_columns
        return index_columns + [column for column in self._data_columns
                                if column not in index_columns]

    @property
    def index_map(self) -> List[IndexMap]:
        """ Return the managed index information. """
        return self._index_map

    @property
    def index_names(self) -> List[Optional[str]]:
        """ Return the managed index names. """
        return [index_name for _, index_name in self._index_map]

    def copy(self, data_columns: Optional[List[str]] = None,
             index_map: Optional[List[IndexMap]] = None) -> 'Metadata':
        """ Copy the metadata.

        :param data_columns: the new column field names. If None, then the original ones are used.
        :param index_map: the new index information. If None, then the original one is used.
        :return: the copied metadata.
        """
        if data_columns is None:
            data_columns = self._data_columns
        if index_map is None:
            index_map = self._index_map
        return Metadata(data_columns=data_columns.copy(), index_map=index_map.copy())

    @staticmethod
    def from_pandas(pdf: pd.DataFrame) -> 'Metadata':
        """ Create a metadata from pandas DataFrame.

        :param pdf: :class:`pd.DataFrame`
        :return: the created metadata
        """
        data_columns = [str(col) for col in pdf.columns]
        index = pdf.index

        index_map = []  # type: List[IndexMap]
        if isinstance(index, pd.MultiIndex):
            if index.names is None:
                index_map = [('__index_level_{}__'.format(i), None)
                             for i in range(len(index.levels))]
            else:
                index_map = [('__index_level_{}__'.format(i) if name is None else name, name)
                             for i, name in enumerate(index.names)]
        else:
            index_map = [(index.name
                          if index.name is not None else '__index_level_0__', index.name)]

        return Metadata(data_columns=data_columns, index_map=index_map)



#
# Copyright (C) 2019 Databricks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
A wrapper for GroupedData to behave similar to pandas GroupBy.
"""

from functools import partial
from typing import Any, List
import numpy as np

from pyspark.sql import functions as F
from pyspark.sql.types import FloatType, DoubleType, NumericType

from databricks import koalas as ks  # For running doctests and reference resolution in PyCharm.
from databricks.koalas.frame import DataFrame
from databricks.koalas.metadata import Metadata
from databricks.koalas.missing.groupby import _MissingPandasLikeDataFrameGroupBy, \
    _MissingPandasLikeSeriesGroupBy
from databricks.koalas.series import Series, _col


class GroupBy(object):
    """
    :ivar _kdf: The parent dataframe that is used to perform the groupby
    :type _kdf: DataFrame
    :ivar _groupkeys: The list of keys that will be used to perform the grouping
    :type _groupkeys: List[Series]
    """

    # TODO: Series support is not implemented yet.
    # TODO: not all arguments are implemented comparing to Pandas' for now.
    def aggregate(self, func_or_funcs, *args, **kwargs):
        """Aggregate using one or more operations over the specified axis.

        Parameters
        ----------
        func : dict
             a dict mapping from column name (string) to aggregate functions (string).

        Returns
        -------
        Series or DataFrame

            The return can be:

            * Series : when DataFrame.agg is called with a single function
            * DataFrame : when DataFrame.agg is called with several functions

            Return Series or DataFrame.

        Notes
        -----
        `agg` is an alias for `aggregate`. Use the alias.

        Examples
        --------

        >>> df = ks.DataFrame({'A': [1, 1, 2, 2],
        ...                    'B': [1, 2, 3, 4],
        ...                    'C': [0.362, 0.227, 1.267, -0.562]},
        ...                   columns=['A', 'B', 'C'])

        >>> df
           A  B      C
        0  1  1  0.362
        1  1  2  0.227
        2  2  3  1.267
        3  2  4 -0.562

        Different aggregations per column

        >>> aggregated = df.groupby('A').agg({'B': 'min', 'C': 'sum'})
        >>> aggregated[['B', 'C']]  # doctest: +NORMALIZE_WHITESPACE
           B      C
        A
        1  1  0.589
        2  3  0.705

        """
        if not isinstance(func_or_funcs, dict) or \
                not all(isinstance(key, str) and isinstance(value, str)
                        for key, value in func_or_funcs.items()):
            raise ValueError("aggs must be a dict mapping from column name (string) to aggregate "
                             "functions (string).")

        sdf = self._kdf._sdf
        groupkeys = self._groupkeys
        groupkey_cols = [s._scol.alias('__index_level_{}__'.format(i))
                         for i, s in enumerate(groupkeys)]
        reordered = [F.expr('{1}({0}) as {0}'.format(key, value))
                     for key, value in func_or_funcs.items()]
        sdf = sdf.groupby(*groupkey_cols).agg(*reordered)
        metadata = Metadata(data_columns=[key for key, _ in func_or_funcs.items()],
                            index_map=[('__index_level_{}__'.format(i), s.name)
                                       for i, s in enumerate(groupkeys)])
        return DataFrame(sdf, metadata)

    agg = aggregate

    def count(self):
        """
        Compute count of group, excluding missing values.

        >>> df = ks.DataFrame({'A': [1, 1, 2, 1, 2],
        ...                    'B': [np.nan, 2, 3, 4, 5],
        ...                    'C': [1, 2, 1, 1, 2]}, columns=['A', 'B', 'C'])
        >>> df.groupby('A').count()  # doctest: +NORMALIZE_WHITESPACE
            B  C
        A
        1  2  3
        2  2  2
        """
        return self._reduce_for_stat_function(F.count, only_numeric=False)

    # TODO: We should fix See Also when Series implementation is finished.
    def first(self):
        """
        Compute first of group values.

        See Also
        --------

        koalas.DataFrame.groupby
        """
        return self._reduce_for_stat_function(F.first, only_numeric=False)

    def last(self):
        """
        Compute last of group values.

        See Also
        --------

        koalas.DataFrame.groupby
        """
        return self._reduce_for_stat_function(lambda col: F.last(col, ignorenulls=True),
                                              only_numeric=False)

    def max(self):
        """
        Compute max of group values.

        See Also
        --------

        koalas.DataFrame.groupby
        """
        return self._reduce_for_stat_function(F.max, only_numeric=False)

    # TODO: examples should be updated.
    def mean(self):
        """
        Compute mean of groups, excluding missing values.

        Returns
        -------
        koalas.Series or koalas.DataFrame

        See Also
        --------

        koalas.DataFrame.groupby

        Examples
        --------
        >>> df = ks.DataFrame({'A': [1, 1, 2, 1, 2],
        ...                    'B': [np.nan, 2, 3, 4, 5],
        ...                    'C': [1, 2, 1, 1, 2]}, columns=['A', 'B', 'C'])

        Groupby one column and return the mean of the remaining columns in
        each group.

        >>> df.groupby('A').mean()  # doctest: +NORMALIZE_WHITESPACE
             B         C
        A
        1  3.0  1.333333
        2  4.0  1.500000
        """

        return self._reduce_for_stat_function(F.mean, only_numeric=True)

    def min(self):
        """
        Compute min of group values.

        See Also
        --------

        koalas.DataFrame.groupby
        """
        return self._reduce_for_stat_function(F.min, only_numeric=False)

    # TODO: sync the doc and implement `ddof`.
    def std(self):
        """
        Compute standard deviation of groups, excluding missing values.
        """

        return self._reduce_for_stat_function(F.stddev, only_numeric=True)

    def sum(self):
        """
        Compute sum of group values

        See Also
        --------

        koalas.DataFrame.groupby
        """
        return self._reduce_for_stat_function(F.sum, only_numeric=True)

    # TODO: sync the doc and implement `ddof`.
    def var(self):
        """
        Compute variance of groups, excluding missing values.
        """
        return self._reduce_for_stat_function(F.variance, only_numeric=True)

    def _reduce_for_stat_function(self, sfun, only_numeric):
        groupkeys = self._groupkeys
        groupkey_cols = [s._scol.alias('__index_level_{}__'.format(i))
                         for i, s in enumerate(groupkeys)]
        sdf = self._kdf._sdf

        data_columns = []
        if len(self._agg_columns) > 0:
            stat_exprs = []
            for ks in self._agg_columns:
                spark_type = ks.spark_type
                # TODO: we should have a function that takes dataframes and converts the numeric
                # types. Converting the NaNs is used in a few places, it should be in utils.
                # Special handle floating point types because Spark's count treats nan as a valid
                # value, whereas Pandas count doesn't include nan.
                if isinstance(spark_type, DoubleType) or isinstance(spark_type, FloatType):
                    stat_exprs.append(sfun(F.nanvl(ks._scol, F.lit(None))).alias(ks.name))
                    data_columns.append(ks.name)
                elif isinstance(spark_type, NumericType) or not only_numeric:
                    stat_exprs.append(sfun(ks._scol).alias(ks.name))
                    data_columns.append(ks.name)
            sdf = sdf.groupby(*groupkey_cols).agg(*stat_exprs)
        else:
            sdf = sdf.select(*groupkey_cols).distinct()
        sdf = sdf.sort(*groupkey_cols)
        metadata = Metadata(data_columns=data_columns,
                            index_map=[('__index_level_{}__'.format(i), s.name)
                                       for i, s in enumerate(groupkeys)])
        return DataFrame(sdf, metadata)


class DataFrameGroupBy(GroupBy):

    def __init__(self, kdf: DataFrame, by: List[Series], agg_columns: List[str] = None):
        self._kdf = kdf
        self._groupkeys = by

        if agg_columns is None:
            groupkey_names = set(s.name for s in self._groupkeys)
            agg_columns = [col for col in self._kdf._metadata.data_columns
                           if col not in groupkey_names]
        self._agg_columns = [kdf[col] for col in agg_columns]

    def __getattr__(self, item: str) -> Any:
        if hasattr(_MissingPandasLikeDataFrameGroupBy, item):
            property_or_func = getattr(_MissingPandasLikeDataFrameGroupBy, item)
            if isinstance(property_or_func, property):
                return property_or_func.fget(self)  # type: ignore
            else:
                return partial(property_or_func, self)
        return self.__getitem__(item)

    def __getitem__(self, item):
        if isinstance(item, str):
            return SeriesGroupBy(self._kdf[item], self._groupkeys)
        else:
            # TODO: check that item is a list of strings
            return DataFrameGroupBy(self._kdf, self._groupkeys, item)


class SeriesGroupBy(GroupBy):

    def __init__(self, ks: Series, by: List[Series]):
        self._ks = ks
        self._groupkeys = by

    def __getattr__(self, item: str) -> Any:
        if hasattr(_MissingPandasLikeSeriesGroupBy, item):
            property_or_func = getattr(_MissingPandasLikeSeriesGroupBy, item)
            if isinstance(property_or_func, property):
                return property_or_func.fget(self)  # type: ignore
            else:
                return partial(property_or_func, self)
        raise AttributeError(item)

    @property
    def _kdf(self) -> DataFrame:
        return self._ks._kdf

    @property
    def _agg_columns(self):
        return [self._ks]

    def _reduce_for_stat_function(self, sfun, only_numeric):
        return _col(super(SeriesGroupBy, self)._reduce_for_stat_function(sfun, only_numeric))



#
# Copyright (C) 2019 Databricks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import pytest
import numpy
import tempfile
import atexit
import shutil
import uuid
from distutils.version import LooseVersion

from pyspark import __version__

from databricks import koalas
from databricks.koalas import utils


# Initialize Spark session that should be used in doctests or unittests.
# Delta requires Spark 2.4.2+. See
# https://github.com/delta-io/delta#compatibility-with-apache-spark-versions.
if LooseVersion(__version__) >= LooseVersion("3.0.0"):
    session = utils.default_session({"spark.jars.packages": "io.delta:delta-core_2.12:0.1.0"})
elif LooseVersion(__version__) >= LooseVersion("2.4.2"):
    session = utils.default_session({"spark.jars.packages": "io.delta:delta-core_2.11:0.1.0"})
else:
    session = utils.default_session()


@pytest.fixture(autouse=True)
def add_ks(doctest_namespace):
    doctest_namespace['ks'] = koalas


@pytest.fixture(autouse=True)
def add_np(doctest_namespace):
    doctest_namespace['np'] = numpy


@pytest.fixture(autouse=True)
def add_path(doctest_namespace):
    path = tempfile.mkdtemp()
    atexit.register(lambda: shutil.rmtree(path, ignore_errors=True))
    doctest_namespace['path'] = path


@pytest.fixture(autouse=True)
def add_db(doctest_namespace):
    db_name = str(uuid.uuid4()).replace("-", "")
    session.sql("CREATE DATABASE %s" % db_name)
    atexit.register(lambda: session.sql("DROP DATABASE IF EXISTS %s CASCADE" % db_name))
    doctest_namespace['db'] = db_name



#
# Copyright (C) 2019 Databricks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
String functions on Koalas Series
"""
from typing import TYPE_CHECKING

import numpy as np

from pyspark.sql.types import StringType, BinaryType, BooleanType

from databricks.koalas.base import _wrap_accessor_pandas

if TYPE_CHECKING:
    import databricks.koalas as ks


class StringMethods(object):
    """String methods for Koalas Series"""
    def __init__(self, series: 'ks.Series'):
        if not isinstance(series.spark_type, (StringType, BinaryType)):
            raise ValueError(
                "Cannot call StringMethods on type {}"
                .format(series.spark_type))
        self._data = series
        self.name = self._data.name

    # Methods
    def capitalize(self) -> 'ks.Series':
        """
        Convert Strings in the series to be capitalized.
        """
        return _wrap_accessor_pandas(
            self,
            lambda x: x.str.capitalize(),
            StringType()
        ).alias(self.name)

    def lower(self) -> 'ks.Series':
        """
        Convert strings in the Series/Index to all lowercase.
        """
        return _wrap_accessor_pandas(
            self,
            lambda x: x.str.lower(),
            StringType()
        ).alias(self.name)

    def upper(self) -> 'ks.Series':
        """
        Convert strings in the Series/Index to all uppercase.
        """
        return _wrap_accessor_pandas(
            self,
            lambda x: x.str.upper(),
            StringType()
        ).alias(self.name)

    def swapcase(self) -> 'ks.Series':
        """
        Convert strings in the Series/Index to be swapcased.
        """
        return _wrap_accessor_pandas(
            self,
            lambda x: x.str.swapcase(),
            StringType()
        ).alias(self.name)

    def startswith(self, pattern, na=np.NaN) -> 'ks.Series':
        """
        Test if the start of each string element matches a pattern.

        Equivalent to :func:`str.startswith`.

        Parameters
        ----------
        pattern : str
            Character sequence. Regular expressions are not accepted.
        na : object, defulat NaN
            Object shown if element is not a string.

        Returns
        -------
        Series of bool
            Koalas Series of booleans indicating whether the given pattern
            matches the start of each string element.
        """
        return _wrap_accessor_pandas(
            self,
            lambda x: x.str.startswith(pattern, na),
            BooleanType()
        ).alias(self.name)

    def endswith(self, pattern, na=np.NaN) -> 'ks.Series':
        """
        Test if the end of each string element matches a pattern.

        Equivalent to :func:`str.endswith`.

        Parameters
        ----------
        pattern : str
            Character sequence. Regular expressions are not accepted.
        na : object, defulat NaN
            Object shown if element is not a string.

        Returns
        -------
        Series of bool
            Koalas Series of booleans indicating whether the given pattern
            matches the end of each string element.
        """
        return _wrap_accessor_pandas(
            self,
            lambda x: x.str.endswith(pattern, na),
            BooleanType()
        ).alias(self.name)

    def strip(self, to_strip=None) -> 'ks.Series':
        """
        Remove leading and trailing characters.

        Strip whitespaces (including newlines) or a set of specified
        characters from each string in the Series/Index from left and
        right sides. Equivalent to :func:`str.strip`.

        Parameters
        ----------
        to_strip : str
            Specifying the set of characters to be removed. All combinations
            of this set of characters will be stripped. If None then
            whitespaces are removed.

        Returns
        -------
        Series of str
        """
        return _wrap_accessor_pandas(
            self,
            lambda x: x.str.strip(to_strip),
            StringType()
        ).alias(self.name)

    def lstrip(self, to_strip=None) -> 'ks.Series':
        """
        Remove leading characters.

        Strip whitespaces (including newlines) or a set of specified
        characters from each string in the Series/Index from left side.
        Equivalent to :func:`str.lstrip`.

        Parameters
        ----------
        to_strip : str
            Specifying the set of characters to be removed. All combinations
            of this set of characters will be stripped. If None then
            whitespaces are removed.

        Returns
        -------
        Series of str
        """
        return _wrap_accessor_pandas(
            self,
            lambda x: x.str.lstrip(to_strip),
            StringType()
        ).alias(self.name)

    def rstrip(self, to_strip=None) -> 'ks.Series':
        """
        Remove trailing characters.

        Strip whitespaces (including newlines) or a set of specified
        characters from each string in the Series/Index from right side.
        Equivalent to :func:`str.rstrip`.

        Parameters
        ----------
        to_strip : str
            Specifying the set of characters to be removed. All combinations
            of this set of characters will be stripped. If None then
            whitespaces are removed.

        Returns
        -------
        Series of str
        """
        return _wrap_accessor_pandas(
            self,
            lambda x: x.str.rstrip(to_strip),
            StringType()
        ).alias(self.name)

    def get(self, i) -> 'ks.Series':
        """
        Extract element from each string in the Series/Index at the
        specified position.

        Parameters
        ----------
        i : int
            Position of element to extract.

        Returns
        -------
        Series of objects
        """
        return _wrap_accessor_pandas(
            self,
            lambda x: x.str.get(i),
            StringType()
        ).alias(self.name)

    def isalnum(self) -> 'ks.Series':
        """
        Check whether all characters in each string are alphanumeric.

        This is equivalent to running the Python string method
        :func:`str.isalnum` for each element of the Series/Index.
        If a string has zero characters, False is returned for that check.
        """
        return _wrap_accessor_pandas(
            self,
            lambda x: x.str.isalnum(),
            BooleanType()
        ).alias(self.name)

    def isalpha(self) -> 'ks.Series':
        """
        Check whether all characters in each string are alphabetic.

        This is equivalent to running the Python string method
        :func:`str.isalpha` for each element of the Series/Index.
        If a string has zero characters, False is returned for that check.
        """
        return _wrap_accessor_pandas(
            self,
            lambda x: x.str.isalpha(),
            BooleanType()
        ).alias(self.name)

    def isdigit(self) -> 'ks.Series':
        """
        Check whether all characters in each string are digits.

        This is equivalent to running the Python string method
        :func:`str.isdigit` for each element of the Series/Index.
        If a string has zero characters, False is returned for that check.
        """
        return _wrap_accessor_pandas(
            self,
            lambda x: x.str.isdigit(),
            BooleanType()
        ).alias(self.name)

    def isspace(self) -> 'ks.Series':
        """
        Check whether all characters in each string are whitespaces.

        This is equivalent to running the Python string method
        :func:`str.isspace` for each element of the Series/Index.
        If a string has zero characters, False is returned for that check.
        """
        return _wrap_accessor_pandas(
            self,
            lambda x: x.str.isspace(),
            BooleanType()
        ).alias(self.name)

    def islower(self) -> 'ks.Series':
        """
        Check whether all characters in each string are lowercase.

        This is equivalent to running the Python string method
        :func:`str.islower` for each element of the Series/Index.
        If a string has zero characters, False is returned for that check.
        """
        return _wrap_accessor_pandas(
            self,
            lambda x: x.str.islower(),
            BooleanType()
        ).alias(self.name)

    def isupper(self) -> 'ks.Series':
        """
        Check whether all characters in each string are uppercase.

        This is equivalent to running the Python string method
        :func:`str.isupper` for each element of the Series/Index.
        If a string has zero characters, False is returned for that check.
        """
        return _wrap_accessor_pandas(
            self,
            lambda x: x.str.isupper(),
            BooleanType()
        ).alias(self.name)

    def istitle(self) -> 'ks.Series':
        """
        Check whether all characters in each string are titlecase.

        This is equivalent to running the Python string method
        :func:`str.istitle` for each element of the Series/Index.
        If a string has zero characters, False is returned for that check.
        """
        return _wrap_accessor_pandas(
            self,
            lambda x: x.str.istitle(),
            BooleanType()
        ).alias(self.name)

    def isnumeric(self) -> 'ks.Series':
        """
        Check whether all characters in each string are numeric.

        This is equivalent to running the Python string method
        :func:`str.isnumeric` for each element of the Series/Index.
        If a string has zero characters, False is returned for that check.
        """
        return _wrap_accessor_pandas(
            self,
            lambda x: x.str.isnumeric(),
            BooleanType()
        ).alias(self.name)

    def isdecimal(self) -> 'ks.Series':
        """
        Check whether all characters in each string are decimals.

        This is equivalent to running the Python string method
        :func:`str.isdecimal` for each element of the Series/Index.
        If a string has zero characters, False is returned for that check.
        """
        return _wrap_accessor_pandas(
            self,
            lambda x: x.str.isdecimal(),
            BooleanType()
        ).alias(self.name)



#
# Copyright (C) 2019 Databricks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
Date/Time related functions on Koalas Series
"""
from typing import TYPE_CHECKING

import pandas as pd
import pyspark.sql.functions as F
from pyspark.sql.types import DateType, TimestampType, LongType, StringType

from databricks.koalas.base import _wrap_accessor_pandas, _wrap_accessor_spark

if TYPE_CHECKING:
    import databricks.koalas as ks


class DatetimeMethods(object):
    """Date/Time methods for Koalas Series"""
    def __init__(self, series: 'ks.Series'):
        if not isinstance(series.spark_type, (DateType, TimestampType)):
            raise ValueError(
                "Cannot call DatetimeMethods on type {}"
                .format(series.spark_type))
        self._data = series
        self.name = self._data.name

    # Properties
    @property
    def date(self) -> 'ks.Series':
        """
        Returns a Series of python datetime.date objects (namely, the date
        part of Timestamps without timezone information).
        """
        # TODO: Hit a weird exception
        # syntax error in attribute name: `to_date(`start_date`)` with alias
        return _wrap_accessor_spark(
            self, lambda col: F.to_date(col)).alias(self.name)

    @property
    def time(self) -> 'ks.Series':
        raise NotImplementedError()

    @property
    def timetz(self) -> 'ks.Series':
        raise NotImplementedError()

    @property
    def year(self) -> 'ks.Series':
        """
        The year of the datetime.
        `"""
        return _wrap_accessor_spark(self, F.year, LongType()).alias(self.name)

    @property
    def month(self) -> 'ks.Series':
        """
        The month of the timestamp as January = 1 December = 12.
        """
        return _wrap_accessor_spark(self, F.month, LongType()).alias(self.name)

    @property
    def day(self) -> 'ks.Series':
        """
        The days of the datetime.
        """
        return _wrap_accessor_spark(
            self, F.dayofmonth, LongType()).alias(self.name)

    @property
    def hour(self) -> 'ks.Series':
        """
        The hours of the datetime.
        """
        return _wrap_accessor_spark(self, F.hour, LongType()).alias(self.name)

    @property
    def minute(self) -> 'ks.Series':
        """
        The minutes of the datetime.
        """
        return _wrap_accessor_spark(self, F.minute, LongType()).alias(self.name)

    @property
    def second(self) -> 'ks.Series':
        """
        The seconds of the datetime.
        """
        return _wrap_accessor_spark(self, F.second, LongType()).alias(self.name)

    @property
    def millisecond(self) -> 'ks.Series':
        """
        The milliseconds of the datetime.
        """
        return _wrap_accessor_pandas(
            self, lambda x: x.dt.millisecond, LongType()).alias(self.name)

    @property
    def microsecond(self) -> 'ks.Series':
        """
        The microseconds of the datetime.
        """
        return _wrap_accessor_pandas(
            self, lambda x: x.dt.microsecond, LongType()).alias(self.name)

    @property
    def nanosecond(self) -> 'ks.Series':
        raise NotImplementedError()

    @property
    def week(self) -> 'ks.Series':
        """
        The week ordinal of the year.
        """
        return _wrap_accessor_spark(self, F.weekofyear, LongType()).alias(self.name)

    @property
    def weekofyear(self) -> 'ks.Series':
        return self.week

    weekofyear.__doc__ = week.__doc__

    @property
    def dayofweek(self) -> 'ks.Series':
        """
        The day of the week with Monday=0, Sunday=6.

        Return the day of the week. It is assumed the week starts on
        Monday, which is denoted by 0 and ends on Sunday which is denoted
        by 6. This method is available on both Series with datetime
        values (using the `dt` accessor) or DatetimeIndex.

        Returns
        -------
        Series or Index
            Containing integers indicating the day number.

        See Also
        --------
        Series.dt.dayofweek : Alias.
        Series.dt.weekday : Alias.
        Series.dt.day_name : Returns the name of the day of the week.

        Examples
        --------
        >>> s = ks.from_pandas(pd.date_range('2016-12-31', '2017-01-08', freq='D').to_series())
        >>> s.dt.dayofweek
        2016-12-31    5
        2017-01-01    6
        2017-01-02    0
        2017-01-03    1
        2017-01-04    2
        2017-01-05    3
        2017-01-06    4
        2017-01-07    5
        2017-01-08    6
        Name: 0, dtype: int64
        """
        return _wrap_accessor_pandas(
            self, lambda s: s.dt.dayofweek, LongType()).alias(self._data.name)

    @property
    def weekday(self) -> 'ks.Series':
        return self.dayofweek

    weekday.__doc__ = dayofweek.__doc__

    @property
    def dayofyear(self) -> 'ks.Series':
        """
        The ordinal day of the year.
        """
        return _wrap_accessor_pandas(
            self, lambda s: s.dt.dayofyear, LongType()).alias(self._data.name)

    # Methods
    def strftime(self, date_format) -> 'ks.Series':
        """
        Convert to a String Series using specified date_format.
        """
        return _wrap_accessor_pandas(
            self,
            lambda x: x.dt.strftime(date_format),
            StringType()
        ).alias(self.name)
